# Developed by Jayme Hebinck, for bachelor thesis Leiden University.
# This file is part of the ADT Create project, found on: https://github.com/MainJay/ADTPythonLibrary/.

from ADTCreate.ADTSatisfiability import Satisfiability
from ADTCreate.GenerateADTerms import GenADTerms
from ADTCreate.CompareADTerms import CompADTerms
from ADTCreate.ADTXML import XML

from typing import List, Tuple, Dict, Union, Optional
from xml.dom import minidom


class ADT:
    """
    Objects of class `ADT` form individual nodes. These nodes are connected to each other using the class variables to
    form an ADT. This class forms the center point of **ADT Create**, and therefore links objects of all classes
    together in this class.
    :param idNumber: Determines the ID number (`int`) of a node. ID numbers enable the introduction of `twin nodes`, meaning the occurances of duplicate, independent nodes.
    :param type: Determines the type of a node (`0` for attack, `1` for defense).
    :param refinement: Determines the refinement of a node (`0` for disjunctive (OR), `1` for conjunctive (AND)).
    :param label: Determines the label of a node.
    :param parent: Determines the parent of a node (`ADT` object, `0` for no parent).
    :param root: Contains two variables in `List`. First element is a `Bool` whether this node is the root node. Second element is an `ADT` object of the root node.
    """
    usedIDs = []

    def __init__(self, idNumber: int, type: int, refinement: int, label: str, parent: Optional['ADT'], root: List[Tuple[bool, Optional['ADT']]]) -> None:
        """
        Constructor of `ADT` class. Initializes values for all membervariables needed.
        :param idNumber: Determines the ID number (`int`) of a node. ID numbers enable the introduction of `twin nodes`, meaning the occurances of duplicate, independent nodes.
        :param type: Determines the type of a node (`0` for attack, `1` for defense).
        :param refinement: Determines the refinement of a node (`0` for disjunctive (OR), `1` for conjunctive (AND)).
        :param label: Determines the label of a node.
        :param parent: Determines the parent of a node (`ADT` object, `0` for no parent).
        :param root: Contains two variables in `List`. First element is a `Bool` whether this node is the root node. Second element is an `ADT` object of the root node.
        """
        # Constructor of class ADT, representing a single ADT node.
        self.children = []
        self.type = type
        self.refinement = refinement
        self.id = idNumber
        self.label = label
        self.parent = parent
        self.root = root
        self.hasCounterChild = False # If a node has one counter child node
        self.leafNode = False # Check whether this node is a leaf node
        self.satisfiability = False # States if the ADT with this node as root node is satisfiable.
        self.ADTerms = "" # ADTerm of ADT witht he current node as root node.
        self.LSEquivalenceThreshold = 0.2 # Levenshtein Distance Equivalence Threshold Value
        self.count = 0 # Counter for how many nodes satisfiability is determined of

        # Class instances:
        self.XML = XML(self)
        self.Satisfiability = Satisfiability(self)
        self.genADTerms = GenADTerms(self)
        self.compADTerms = CompADTerms(self)

    def addChild(self, typeChild: int, refinementChild: int, labelChild: str, currentChild: 'ADT', tree: 'ADT') -> Tuple[bool, Optional['ADT']]:
        """
        Adds a child node with the given parameters as new node to the `tree`, which is the root of the ADT.
        Child node is given its own unique ID.

        :param typeChild: Type of the child node (`0` for attack, `1` for defense).
        :param refinementChild: Refinement of the child node (`0` for disjunctive (OR), `1` for conjunctive (AND)).
        :param labelChild: Label of the child node.
        :param currentChild: `ADT` object of the parent node, to which this child node is connected.
        :param tree: Root node of the ADT this node is connected to.
        :return:`bool`, `ADT`/`None`: `True` and `ADT` object of new child if child is added successfully, `False` and `None` otherwise.
        """
        idNumber = self.usedIDs[-1] + 1 # Take latest ID and increment by one.
        newChild = ADT(idNumber, int(typeChild), int(refinementChild), labelChild, currentChild, [False, tree])
        # Firstly check if this child can be added in the boundaries of Countermeasure Prevention.
        if not self.checkCounterMeasure(newChild):
            return False, None
        # Set this child as leaf node, and if parent node was leaf node, it is not anymore.
        newChild.leafNode = True
        self.children.append(newChild)
        if self.leafNode is True:
            self.leafNode = False
        self.usedIDs.append(idNumber) # Add latest ID to this list.
        return True, newChild

    def delChild(self, id: int) -> bool:
        """
        Deletes a child node with the given ID.

        :param id: ID of child node.
        :return: `bool`: True if child node is deleted successfully, False otherwise.
        """
        # ID of child first has to be found in IDs of current children.
        found = False
        for child in self.children:
            if child.id == int(id):
                found = True
                childToRemove = child
        if not found:
            return False
        # If found, remove it. Set parent node to leaf node if no other children.
        self.children.remove(childToRemove)
        if int(id) in self.usedIDs:
            self.usedIDs.remove(int(id)) # Remove this ID from usedIDs list.
        if len(self.children) == 0:
            self.leafNode = True
        # If this was a countermeasure node, turn the countermeasure variable off.
        if (self.type == 0 and self.hasCounterChild and child.type == 1) or (self.type == 1 and self.hasCounterChild and child.type == 0):
            self.hasCounterChild = False
        return True

    def showChildren(self) -> None:
        """
        Shows all children nodes of current node, with their ID. Function only shows children on Terminal.

        :return:
        """
        # Display all the children labels and IDs of current node.
        print("Current node:")
        print(self.label, "(with ID:", self.id, ")")
        print("")
        print("Children node(s):")
        for child in self.children:
            if child.leafNode:
                print(child.label, " (*)", "(with ID:", child.id, ")")
            else:
                print(child.label, "(with ID:", child.id, ")")
        if self.leafNode:
            # If no children, print the following.
            print("[No Children]")

    def changeRefinement(self) -> None:
        """
        Changes refinement of current node. No returns, since this operations cannot lead to an error.

        :return:
        """
        # Change refinement of current node.
        if self.refinement == 0:
            self.refinement = 1
        elif self.refinement == 1:
            self.refinement = 0

    def getNodeInformation(self) -> List[Union[Tuple[str, int, int, list, str, int], Tuple[str, int, int, list, None, int]]]:
        """
        Retrieve label, type, refinement, children, parent and ID in a list.

        :return: `list`: Contains label, type, refinement, children, parent and ID respectively.
        """
        # Retrieve label, type, refinement, parent and ID in a list.
        childLabels = []
        for i in range(len(self.children)):
            childLabels.append(self.children[i].label)
        if self.parent != None:
            return self.label, self.type, self.refinement, childLabels, self.parent.label, self.id
        else:
            return self.label, self.type, self.refinement, childLabels, None, self.id

    def changeType(self) -> bool:
        """
        Changes type of current node.

        :return: `bool`: `True` when successful, `False` otherwise.
        """
        # If current type is attack, change to defense. If current type is defense, change to attack.
        if self.type == 0:
            self.type = 1
        else:
            self.type = 0
        oneCounterChild = False # Parent is allowed to have at most one counter child.
        for child in self.children:
            # For all children, check if current child has counter. If not, set oneCounterChild to True. If yes, countermeasure prevention
            # prohibits from this node changing type.
            if self.type == 0:
                if child.type == 1 and not oneCounterChild:
                    oneCounterChild = True
                elif child.type == 1 and oneCounterChild:
                    if self.type == 0:
                        self.type = 1
                    else:
                        self.type = 0
                    return False
            else:
                if child.type == 0 and not oneCounterChild:
                    oneCounterChild = True
                elif child.type == 0 and oneCounterChild:
                    if self.type == 0:
                        self.type = 1
                    else:
                        self.type = 0
                    return False
        # If type is changed successfully, clear full countermeasure prevention and reset it all in the tree so that
        # this type change is registered in the total ADT countermeasure situation.
        self.root[1].clearCountermeasurePrevention()
        self.root[1].checkAllCountermeasurePrevention()
        return True

    def changeLabel(self, newLabel: str) -> bool:
        """
        Changes label of current node to provided label.

        :param newLabel: Contains the new label of the node.
        :return: `bool`: `True` when successful, `False` otherwise.
        """
        # Try changing the label. If something goes wrong, return False.
        try:
            self.label = newLabel
            return True
        except:
            return False

    def currentNodeToChildNode(self, idChild: int) -> Union['ADT', bool]:
        """
        Changes the current node that can be worked on to the child node with the given ID.

        :param idChild: `idChild`: ID of the child node.
        :return: `ADT`/`None`: `ADT` object of child node when child node with given ID exists, `None` otherwise.
        """
        # Check if given ID is indeed in the current node's children.
        try:
            for child in self.children:
                if int(idChild) == child.id:
                    # If yes, return this child.
                    return child
        except:
            # If not, return False.
            return False

    def currentNodeToParentNode(self) -> Union['ADT', bool]:
        """
        Changes the current node that can be worked on to the parent node.

        :return: `ADT`/`None`: `ADT` object of parent node when parent node is available, `None` if node has no parent (f.e. root node).
        """
        # If we are not in the root node, go to the parent node. Otherwise, return False.
        if (self.parent != 0):
            return self.parent
        else:
            return False


    def checkCounterMeasure(self, child: 'ADT') -> bool:
        """
        Checks if the countermeasure prevention rules are kept.

        :param child: Child between which and the parent node countermeasure prevention must be checked.
        :return: `bool`: `True` if countermeasure prevention rules are kept, `False` otherwise.

        Note: *Helper function, users should not call this function themselves. Used in member functions: `addChild(...)` and `clearCountermeasurePrevention(...)`.*
        """
        # If this node already has a counter, and new child is also a counter child, prohibit action and return False.
        if (self.hasCounterChild and self.type == 1 and child.type == 0) or (self.hasCounterChild and self.type == 0 and child.type == 1):
            return False
        # If this node does not have a counter child yet, set counter to True.
        if (self.type == 0 and child.type == 1) or (self.type == 1 and child.type == 0):
            self.hasCounterChild = True
        # Countermeasure prevention is checked! Return True.
        return True

    def clearCountermeasurePrevention(self) -> None:
        """
        Clears Countermeasure Prevention. Resets `hasCounterChild` variables of all nodes in ADT to `False`.
        :return:

        Note: *Helper function, users should not call this function themselves. Used in member function: `checkAllCountermeasurePrevention(self)`.*
        """
        # For all children in Depth-First Search, recursively call this function. Set counter variables to False.
        for i in range(len(self.children)):
            if len(self.children) > 0:
                self.children[i].clearCountermeasurePrevention()
                self.hasCounterChild = False

    def checkAllCountermeasurePrevention(self) -> bool:
        """
        Checks Countermeasure Prevention after uploading a new ADT in XML format, removes it and redetermines it. Needs to be done manually to prevent Countermeasure Prevention glitches.

        :return: `bool`: `True` if successful, `False` otherwise.
        """
        # For all children in Depth-First Search, recursively call this function.
        for i in range(len(self.children)):
            if len(self.children) > 0:
                if not self.children[i].checkAllCountermeasurePrevention():
                    return False
                if not self.checkCounterMeasure(self.children[i]):
                    # If Countermeasure Prevention prohibits action, return False.
                    return False
        # Countermeasure Prevention over whole tree is successful! Return True.
        return True

    # XML Operations - ADTXML.py
    def importXML(self, fileName: str = "", manual: bool = True) -> Tuple[Union['ADT', bool], int]:
        """
        Imports a new ADT with `self` as the root node `ADT` object in **XML class**. Filename can either be given as parameter or interactively via Terminal (to allow for dynamical file import) and should **not contain .xml**.

        :param fileName: Filename of the XML file, **excluding** .xml.
        :param manual: `True` if you want to give filename interactively, `False` if you want the process to go automatic, without human input needed.
        :return: `ADT`/`bool`, `int`: `ADT` object and `0` of root node of ADT if successful, `False` and `1` if file not found, `False` and `2` if imported ADT does not comply with countermeasure prevention.
        """
        XMLResult = self.XML.importXML(fileName, manual)
        if not XMLResult:
            return False, 1
        elif not XMLResult.checkAllCountermeasurePrevention():
            return False, 2
        return XMLResult, 0

    def exportXML(self, fileName: str = "", manual: bool = True) -> bool:
        """
        Exports the current ADT with `self` as the root node `ADT` object in **XML class**. User has to interactively give a filename for the exported file in the terminal.

        :param fileName: Filename of the XML file, **excluding** .xml.
        :param manual: `True` if you want to give filename interactively, `False` if you want the process to go automatic, without human input needed.
        :return: `bool`: `True` if exporting was successful, `False` otherwise.
        """
        if not self.XML.exportXML(fileName, manual):
            return False
        return True

    def exportChild(self, root: minidom.Document, nodeChild: minidom.Element) -> bool:
        """
        Recursively adds children to the XML structure.

        :param root: Root node, `Document` object of the ADT that is exported.
        :param nodeChild: Children of this node, `ADT` objects, will be appended to this `Element` object.
        :return: `bool`: `True` when exporting children concluded successfully, `False` otherwise.

        Note: *Helper function, users should not call this function themselves. Used in member function: `exportXML(...)`.*
        """
        if self.XML.exportChild(root, nodeChild):
            return True
        return False

    def determineLeafNodes(self) -> None:
        """
        Determines all the leaf nodes in the current ADT structure by setting the variable `self.leafNode` of all `ADT` objects of leaf nodes to `True`. No returns.

        :return:

        Note: *Helper function, users should not call this function themselves. Used in member function: `importXML(...)`.*
        """
        # Go through all nodes recursively.
        for i in range(len(self.children)):
            if len(self.children) > 0:
                self.children[i].determineLeafNodes()
        if not self.children:
            # If this node has no children, then this is a leaf node.
            self.leafNode = True

    # ADT Satisfiability - ADTSatisfiability.py
    def determineSatisfiability(self) -> bool:
        """
        Determines satisfiability of the ADT, with the node it is run on as the root node.

        :return: `bool`: `True` if ADT is satisfiable, `False` if ADT is not satisfiable.
        """
        satResult = self.Satisfiability.determineSatisfiability()
        if satResult:
            self.root[1].count += 1
            return True
        else:
            self.root[1].count += 1
            return False

    # Generate ADTerms - GenerateADTerms.py
    def determineADTerms(self) -> str:
        """
        Determines ADTerms of the ADT, with the node it is run on as the root node.

        :return: `str`: ADTerms of the ADT in string format.
        """
        return self.genADTerms.determineADTerms()

    # Compare ADTerms - CompareADTerms.py
    def compareADTerms(self, ADTerms1: str, ADTerms2: str) -> Union[Tuple[bool, int], Tuple[bool, None]]:
        """
        Determines whether two ADTerms are equivalent to each other by comparing the two. Includes check with equivalence threshold using the Levenshtein Distance.

        :param ADTerms1: ADTerm to compare with, generated by `determineADTerms(...)`.
        :param ADTerms2: ADTerm to compare with, generated by `determineADTerms(...)`.
        :return: `bool`, `int`/`None`: `True` and total Levenshtein Distance (`Int`) if equivalent, `False` and `None` otherwise.
        """
        resultComp = self.compADTerms.compareADTerms(ADTerms1, ADTerms2)
        if not resultComp[0]:
            return False, None
        return True, resultComp[1]

    def changeLevenshteinErrorMargin(self, newValue: float) -> bool:
        """
        Changes Levenshtein Distance equivalence threshold value if new value is between 0.0 and 1.0.

        :param newValue: New value of Levenshtein Distance equivalence threshold.
        :return: `bool`: `True` if value changed successfully, `False` otherwise.
        """
        if not self.compADTerms.changeLevenshteinErrorMargin(newValue):
            return False
        return True


    def generateStatistics(self, attackNodes: int = 0, defenseNodes: int = 0, leafNodes: int = 0, ORRefinements: int = 0, ANDRefinements: int = 0, counterMeasures: int = 0) -> Tuple[int, int, int, int, int, int]:
        """
        Generates statistics of ADT, including the amount of attack nodes, the amount of defense nodes, the amount of leaf nodes, the amount of OR refinements, the amount of AND refinements and the amount of countermeasures.

        :param attackNodes: Amount of attack nodes. Should not be used by user, unless a custom amount should be added. Is necessary for the recursion.
        :param defenseNodes: Amount of defense nodes. Should not be used by user, unless a custom amount should be added. Is necessary for the recursion.
        :param leafNodes: Amount of leaf nodes. Should not be used by user, unless a custom amount should be added. Is necessary for the recursion.
        :param ORRefinements: Amount of OR refinements. Should not be used by user, unless a custom amount should be added. Is necessary for the recursion.
        :param ANDRefinements: Amount of AND refinements. Should not be used by user, unless a custom amount should be added. Is necessary for the recursion.
        :param counterMeasures: Amount of countermeasures. Should not be used by user, unless a custom amount should be added. Is necessary for the recursion.
        :return: `list`: `int` values of the amount of attack nodes, defense nodes, leaf nodes, OR refinements, AND refinements and countermeasures respectively.
        """
        # All statistics are given as arguments of this function (default value is 0, function can be called without all those arguments).
        for i in range(len(self.children)):
            # For the children of the current node.
            if len(self.children) > 0:
                # If this node has children.
                # Go into this function recursively and let the return be all the statistics. We give the currently known statistics to the children (this way, in Depth First Search, f.e. the second child already has the statistics of the first child saved).
                attackNodes, defenseNodes, leafNodes, ORRefinements, ANDRefinements, counterMeasures = self.children[i].generateStatistics(attackNodes, defenseNodes, leafNodes, ORRefinements, ANDRefinements, counterMeasures)
                # From this point on, all statistics are collected Depth-First Search.

                # Determine type of the current node.
                if self.children[i].type == 0:
                    attackNodes += 1
                else:
                    defenseNodes += 1

                # Determine refinement of the current node (if this node is not a leafnode and has more than 1 child).
                if not self.children[i].leafNode and len(self.children[i].children) > 1:
                    if self.children[i].refinement == 0:
                        ORRefinements += 1
                    else:
                        ANDRefinements += 1

                # Determine leaf nodes.
                elif self.children[i].leafNode:
                    leafNodes += 1

                # Determine counter measure.
                if self.children[i].type != self.type:
                    counterMeasures += 1

        # For the root node.
        if self.root[0]:
            # Determine type of root node.
            if self.type == 0:
                attackNodes += 1
            else:
                defenseNodes += 1

            # Determine refinement of root node (if root node has more than 1 child).
            if self.refinement == 0 and len(self.children) > 1:
                ORRefinements += 1
            elif self.refinement == 1 and len(self.children) > 1:
                ANDRefinements += 1

            # Determine if root node has Counter Measure.
            for j in range(len(self.children)):
                if self.children[j].type != self.type:
                    counterMeasures += 1

        # Return current statistics. These statistics are given to (and updated by) the parent node.
        return attackNodes, defenseNodes, leafNodes, ORRefinements, ANDRefinements, counterMeasures

# Function initADT creates an ADT Root node and returns it to the user. This is not a member function of the ADT class.
def initADT(type: int, refinement: int, label: str) -> 'ADT':
    """
    Initializes an ADT with as root node a node with the given type, refinement and label by calling the constructor of
    class `ADT`. Returns the root node to the user, on which memberfunctions of the class `ADT` can be performed. Since
    this function is located in `ADT.py`, the location of this function must be given when calling it. Therefore, the
    prefix `ADT` when calling the function is required.
    :param type: Type of the root node (`0` is attack, `1` is defense).
    :param refinement: Refinement of the root node (`0` is disjunctive (OR), `1` is conjunctive (AND)).
    :param label: Label of the root node.
    :return: `ADT` object: Root node of the newly created ADT.
    """
    id = 0 # Initialize Root node with ID 0.
    parent = None # Root node does not have a parent.
    root = [True, None]
    rootNode = ADT(id, type, refinement, label, parent, root) # Create Root node.
    rootNode.root[1] = rootNode # Set newly created node as Root node.
    rootNode.usedIDs.append(id) # Append this ID to usedID list.
    rootNode.leafNode = True
    return rootNode







# Developed by Jayme Hebinck, for bachelor thesis Leiden University.
# This file is part of the ADT Create project, found on: https://github.com/MainJay/ADTPythonLibrary/

from . import ADT

class GenADTerms:
    """
    Contains functions regarding ADTerms Generation operations on `ADT` objects.
    """
    def __init__(self, ADT: 'ADT.ADT') -> None:
        """
        Links `GenADTerms` object to `ADT` object. `GenADTerms` object is initialized when `ADT` object is initialized.
        :param ADT: Links `ADT` object to this `GenADTerms` object, to make sure these two objects are linked correctly.
        """
        self.ADT = ADT

    def ADTermsIntermediary(self, result: str) -> str:
        """
        Determines ADTerm of an intermediary node (no counter child node) in the ADT.

        :param result: Current ADTerm of children nodes.
        :return:`str`: DTerm of intermediary node.

        Note: *Helper function, users should not call this function themselves. Used in member function: `determineADTerms(...)`.*
        """
        newRes = ""  # Empty string
        # Add refinement and type to string. Proponent will be the type of the root node, Opponent the counternode.
        if self.ADT.type == self.ADT.root[1].type:
            if self.ADT.refinement == 0:
                newRes += '∨p['
            else:
                newRes += '∧p['
        else:
            if self.ADT.refinement == 0:
                newRes += '∨o['
            else:
                newRes += '∧o['
        # Add children labels between brackets.
        newRes += f'{result}]'
        return newRes

    def ADTermsIntermediaryWithCounter(self, counterChild: 'ADT.ADT') -> str:
        """
        Determines ADTerm of an intermediary node with counter child node in the ADT.

        :param counterChild: ADTerm of counter child node.
        :return: `str`: ADTerm of intermediary node.

        Note: *Helper function, users should not call this function themselves. Used in member function: `determineADTerms(...)`.*
        """
        # Reasoning why the current result is NOT being used in this function:
        # Result is already a string with all subresults combined up until this point.
        # The problem is, in this function (only this one) this subresult must be altered, and is not just concatenated
        # to other parts of the string. Therefore, the choice was made to remake this subresult using the saved ADTerms
        # in the nodes of the ADT.
        # This is the most effective way for this function, and using the subresult is the most effective way for
        # other functions. Therefore, this choice has been made.

        newRes = ""  # Empty string
        # Add counter symbol, refinement and type to string
        if self.ADT.type == self.ADT.root[1].type:
            newRes += 'cp['
            if self.ADT.refinement == 0:
                newRes += '∨p['
            else:
                newRes += '∧p['
        else:
            newRes += 'co['
            if self.ADT.refinement == 0:
                newRes += '∨o['
            else:
                newRes += '∧o['

        combinedADTerms = ""  # New empty string to determine belonging children and counter children.
        for thisChild in self.ADT.children:
            if thisChild is not counterChild:
                # For every child, check if this child is not the defense child.
                if combinedADTerms == "":
                    # If first belonging child, do not add any commas.
                    combinedADTerms += thisChild.ADTerms
                else:
                    # If other belong child, add child using a comma before.
                    combinedADTerms += ", " + thisChild.ADTerms
        # Firstly add all belonging children and close this section using ']', then add counter child and close this
        # section using ']' as well.
        combinedADTerms += "], " + counterChild.ADTerms + "]"
        newRes += combinedADTerms  # Concatenate the children to the determined symbols and return.
        return newRes


    def ADTermsSingleNodeWithCounter(self, result: str) -> str:
        """
        Determines ADTerm of an intermediary node (no counter child node) in the ADT.

        :param result: Current ADTerm of children nodes.
        :return: `str`: ADTerm of intermediary node.

        Note: *Helper function, users should not call this function themselves. Used in member function: `determineADTerms(...)`.*
        """
        newRes = ""  # Emptry string
        # Add counter symbol, refinement is not needed.
        if self.ADT.type == self.ADT.root[1].type:
            newRes += 'cp['
        else:
            newRes += 'co['
        # Add first the parent label, then the child label to the ADTerms string.
        newRes += f'"{self.ADT.label}"'
        newRes += f', {result}]'
        return newRes


    def determineADTerms(self) -> str:
        """
        Determines ADTerm of the ADT, with the node it is run on as the root node.

        :return: `str`: ADTerm of the ADT in string format.

        Note: *This function is called from the `ADT` class using `determineADTerms(...)`, and should therefore not be called from the `GenADTerms` class.*
        """
        # Create empty string
        result = ""
        for i in range(len(self.ADT.children)):
            # Save subRes determined from children
            subRes = self.ADT.children[i].determineADTerms()
            if i == 0:
                # If first element, then no comma in between. Otherwise, comma in between.
                result += f'{subRes}'
            else:
                result += f', {subRes}'
        if self.ADT.leafNode:
            # if self.type == self.parent.type:
            result += f'"{self.ADT.label}"'  # Get ADTerms of a leaf node.
        else:
            # If this node is not a leaf node:
            if len(self.ADT.children) == 1:
                # If this node has only one child.
                if self.ADT.children[0].type == self.ADT.type:
                    # Add brackets in between the child to show it is a level deeper (if child is of same type).
                    result = "[" + result + "]"
                else:
                    # Otherwise there is one child with the opposite type, get ADTerms.
                    result = self.ADTermsSingleNodeWithCounter(result)
            else:
                for child in self.ADT.children:
                    # Otherwise, if there are > 1 children, then check for type.
                    if child.type is not self.ADT.type:
                        # If one child of opposite type, get ADTerms.
                        result = self.ADTermsIntermediaryWithCounter(child)
                        break
                else:
                    # Otherwise, if there are > 1 children of the same type, get ADTerms.
                    result = self.ADTermsIntermediary(result)
        self.ADT.ADTerms = result  # Save current ADTerms in the current node.
        return result



# Developed by Jayme Hebinck, for bachelor thesis Leiden University.
# This file is part of the ADT Create project, found on: https://github.com/MainJay/ADTPythonLibrary/

from . import ADT

class Satisfiability:
    """
    Contains functions regarding Satisfiability operations on `ADT` objects.
    """
    def __init__(self, ADT: 'ADT.ADT') -> None:
        """
        Links `Satisfiability` object to `ADT` object. `Satisfiability` object is initialized when `ADT` object is initialized.
        :param ADT: Links `ADT` object to this `Satisfiability` object, to make sure these two objects are linked correctly.
        """
        self.ADT = ADT

    def determineSatisfiability(self) -> bool:
        """
        Determines satisfiability of the ADT, with the node is it run on as the root node.

        :return: `bool`: `True` if ADT is satisfiable, `False` if ADT is not satisfiable.

        Note: *This function is called from the `ADT` class using `determineSatisfiability(...)`, and should therefore not be called from the `Satisfiability` class.*
        """
        if not self.ADT.leafNode and len(self.ADT.children) > 1:
            # For all the nodes that are not leaf nodes and have more than one child.
            if self.ADT.refinement == 0:
                # For every child that has OR refinement.
                for i in range(len(self.ADT.children)):
                    # Either of these options is enough to prove there is at least one satisfiable path.
                    # Check if any direct counter child node for this node.
                    if self.ADT.type == self.ADT.children[i].type:
                        if self.ADT.children[i].determineSatisfiability():
                            for j in range(len(self.ADT.children)):
                                # print(self.ADT.children[j].label)
                                # If a direct counter child node is found, return False
                                if self.ADT.children[j].type != self.ADT.type:
                                    # print(self.ADT.children[j].label) # THIS PART IS GOOD
                                    if self.ADT.children[j].determineSatisfiability():
                                        return False
                            # If the type of the parent and child are the same, and the child is satisfiable, and there is
                            # no direct counter child node, then the parent is also satisfiable.
                            return True
                    elif self.ADT.type != self.ADT.children[i].type:
                        if not self.ADT.children[i].determineSatisfiability():
                            # If the child is a counter node, and it is not satisfiable, then the parent node is
                            # satisfiable.
                            return True
                return False
            else:
                # For every child that has AND refinement.
                for i in range(len(self.ADT.children)):
                    # Either of these options is enough to prove this node is not satisfiable.
                    # Skip the other children, not necessary to check those.
                    if not self.ADT.children[i].determineSatisfiability():
                        # If the child node is not satisfiable, then the parent is not satisfiable.
                        return False
                    elif self.ADT.type != self.ADT.children[i].type:
                        # If the child node is satisfiable, but the child is a counter node, then the parent is not
                        # satisfiable.
                        return False
                return True
        elif not self.ADT.leafNode and len(self.ADT.children) == 1:
            # For all the nodes that are not leaf nodes and have one child.
            if self.ADT.type == self.ADT.children[0].type:
                # If the parent and child nodes are of same type, and child is satisfiable, then so is the parent.
                if self.ADT.children[0].determineSatisfiability():
                    return True
            else:
                # If the parent and child nodes are not of the same type, and the child is not satisfiable, then the
                # parent is.
                # print(self.ADT.label, self.ADT.children[0].label)
                if not self.ADT.children[0].determineSatisfiability():
                    return True
            return False
        else:
            # We assume all basic actions (leaf nodes) are satisfiable.
            return True


# Developed by Jayme Hebinck, for bachelor thesis Leiden University.
# This file is part of the ADT Create project, found on: https://github.com/MainJay/ADTPythonLibrary/.

from xml.dom import minidom
from . import ADT

from typing import List, Tuple, Dict, Union, Optional

class XML:
    """
    Contains functions regarding XML operations on `ADT` objects.
    """
    def __init__(self, ADT: 'ADT.ADT') -> None:
        """
        Links `XML` object to `ADT` object. `XML` object is initialized when `ADT` object is initialized.
        :param ADT: Links `ADT` object to this `XML` object, to make sure these two objects are linked correctly.
        """
        self.ADT = ADT

    def exportChild(self, root: minidom.Document, nodeChild: minidom.Element) -> bool:
        """
        Recursively adds children to the XML structure.

        :param root: Root node, `Document` object of the ADT that is exported.
        :param nodeChild: Children of this node, `ADT` objects, will be appended to this `Element` object.
        :return: `bool`: `True` when exporting children concluded successfully, `False` otherwise.

        Note: *Helper function, users should not call this function themselves. Used in member function: `exportXML(...)`.*

        Note: *This function is called from the `ADT` class using `exportXML(...)`, and should therefore not be called from the `ADTXML` class.*
        """
        # For all children recursively, add the nodes to the .xml.
        for i in range(len(self.ADT.children)):
            nodeChild2 = root.createElement('node')
            # Add refinement and type.
            if (self.ADT.children[i].refinement == 0):
                nodeChild2.setAttribute('refinement', 'disjunctive')
            else:
                nodeChild2.setAttribute('refinement', 'conjunctive')
            if (self.ADT.children[i].type == 1):
                nodeChild2.setAttribute('switchRole', 'yes')
            labelChild2 = root.createElement('label')
            labelChildText2 = root.createTextNode(self.ADT.children[i].label)
            # Add the new child to the existing .xml.
            nodeChild.appendChild(nodeChild2)
            nodeChild2.appendChild(labelChild2)
            labelChild2.appendChild(labelChildText2)
            self.ADT.children[i].exportChild(root, nodeChild2) # Recursively call with next node.
        return True

    def exportXML(self, filename: str = "", manual: bool = True) -> bool:
        """
        Exports the current ADT with `self` as the root node `ADT` object in **XML class**. User has to interactively give a filename for the exported file in the terminal.

        :param filename: Filename of the XML file, **excluding** .xml.
        :param manual: `True` if you want to give filename interactively, `False` if you want the process to go automatic, without human input needed.
        :return: `bool`: `True` if exporting was successful, `False` otherwise.

        Note: *This function is called from the `ADT` class using `exportXML(...)`, and should therefore not be called from the `ADTXML` class.*
        """
        # Initialize needed variables.
        root = minidom.Document()
        xml = root.createElement('adtree')
        root.appendChild(xml)
        nodeChild = root.createElement('node')
        # For the root node, set refinement and type for in the .xml.
        if (self.ADT.refinement == 0):
            nodeChild.setAttribute('refinement', 'disjunctive')
        else:
            nodeChild.setAttribute('refinement', 'conjunctive')
        if (self.ADT.type == 1):
            nodeChild.setAttribute('switchRole', 'yes')
        labelChild = root.createElement('label')
        labelChildText = root.createTextNode(self.ADT.label)
        # Append this to the .xml.
        xml.appendChild(nodeChild)
        nodeChild.appendChild(labelChild)
        labelChild.appendChild(labelChildText)
        # Now with the known .xml structure and the root already in the .xml, add the rest of the nodes to the .xml.
        self.ADT.exportChild(root, nodeChild)

        # Set .xml settings and ask user for output name.
        xmlContent = root.toprettyxml(indent="  ")
        if manual:
            filename = input("File output name (do not add an extension): ")
        filename = filename + ".xml"

        # Write file with given name to the file directory.
        with open(filename, "w") as f:
            f.write(xmlContent)

        return True


    def importXML(self, fileName: str = "", manual: bool = True) -> Union['ADT.ADT', bool]:
        """
        Imports a new ADT with `self` as the root node `ADT` object. Filename can either be given as parameter or interactively via Terminal (to allow for dynamical file import) and should **not contain .xml**.

        :param fileName: Filename of the XML file, **excluding** .xml.
        :param manual: `True` if you want to give filename interactively, `False` if you want the process to go automatic, without human input needed.
        :return: `ADT`/`bool`: `ADT` object of root node of ADT if successful, `False` otherwise.

        Note: *This function is called from the `ADT` class using `importXML(...)`, and should therefore not be called from the `ADTXML` class.*
        """
        if manual:
            fileName = input("File input name (do not add an extension): ")
        try:
            pastusedIDs = self.ADT.usedIDs
            self.ADT.usedIDs.clear() # Resets used IDs if ADTs have been uploaded prior to this ADT.
            root = minidom.parse(fileName + ".xml")  # Import the .xml file
            adtree = root.firstChild  # Gain the first tag (adtree)
            children = adtree.getElementsByTagName("node")  # Retrieve all nodes in the .xml
            labels = adtree.getElementsByTagName("label")  # Retrieve all labels in the .xml
            depths = {}  # Initialize depths dictionary
            treeNodeStructure = {}  # Initialize treeNodeStructure dictionary
            treeLabelStructure = {}  # Initialize treeLabelStructure dictionary
            # Initialize needed variables for looping and identifying nodes.
            if not self.ADT.usedIDs:
                nodeID = 0
            else:
                nodeID = self.ADT.usedIDs[-1] + 1
            treeLoop = 0
            labelLoop = nodeID

            # Retrieve labels
            for label in labels:
                # Save labels to treeLabelStructure dictionary.
                treeLabelStructure[labelLoop] = label.firstChild.nodeValue
                labelLoop = labelLoop + 1

            # Retrieve depth
            depthLoopInit = 0
            deepest = 0
            for child in children:
                # Determine the depth of the nodes and save it in the depths dictionary.
                depth = 0
                while child.parentNode:
                    depth = depth + 1
                    child = child.parentNode
                    if depth > deepest:
                        deepest = depth
                depths[depthLoopInit] = depth
                depthLoopInit = depthLoopInit + 1

            depthParent = [0] * (deepest + 1)  # Initialize list of most recent parents of nodes on all depth levels in .xml

            depthLoop = 0  # Initialize new loop variable

            for child in children:
                # For all the nodes in this .xml
                # Retrieve refinement
                refNodeStr = child.getAttribute("refinement")
                if (refNodeStr == "disjunctive"):
                    # If refinement is set to 'disjunctive', this node has the OR refinement, which is type 0.
                    refNode = 0
                elif (refNodeStr == "conjunctive"):
                    # If refinement is set to 'conjunctive', this node has the AND refinement, which is type 1.
                    refNode = 1

                # Retrieve type
                typeNodeStr = child.getAttribute("switchRole")
                # If switchRole is present in the .xml, this node is of type 1 (defense node)
                if (typeNodeStr == "yes"):
                    typeNode = 1
                else:
                    typeNode = 0

                # Determine parent
                parentNode = 0  # The parent node, used in creating the ADT resulting from the import.
                if depthLoop != 0:
                    # As long as this is not the root.
                    if depths[depthLoop - 1] < depths[depthLoop]:
                        # If depth of previous node is lower than the current node, it is sure that the previous node
                        # is the parent node.
                        depthParent[depths[depthLoop]] = treeNodeStructure[depthLoop - 1]
                        parentNode = treeNodeStructure[depthLoop - 1]
                    elif depths[depthLoop - 1] == depths[depthLoop]:
                        # If depth of previous node is equal to the current node, it is sure that the previous node
                        # is the parent node, but the most recent parent's node depth of this node.
                        parentNode = treeNodeStructure[depthLoop - 1].parent
                    elif depths[depthLoop - 1] > depths[depthLoop]:
                        # If depth of previous node is higher than current node, we go back to the most recent node
                        # on this depth. The parent of this node is guaranteed also the parent of this node.
                        parentNode = depthParent[depths[depthLoop]]

                # Using all gathered information, create ADT of this node.
                ADTInstance = createADTInstance(nodeID, typeNode, refNode, treeLabelStructure[nodeID], parentNode,
                                       [False, None])
                ADTInstance.usedIDs.append(nodeID) # Fill usedIDs list to keep track of all used IDs.
                treeNodeStructure[treeLoop] = ADTInstance


                # Increment all needed variables for looping and identification.
                nodeID = nodeID + 1
                treeLoop = treeLoop + 1
                depthLoop = depthLoop + 1

            for node in range(treeLoop):
                # Create the list of children of every node in the ADTs.
                if (treeNodeStructure[node].parent != 0):
                    treeNodeStructure[node].parent.children.append(treeNodeStructure[node])
                    treeNodeStructure[node].root = (False, treeNodeStructure[0])
                else:
                    treeNodeStructure[node].parent = None # None to describe a node does not have a parent, instead of 0.
                    treeNodeStructure[node].root = [True, treeNodeStructure[0]]
            treeNodeStructure[0].determineLeafNodes()
            return treeNodeStructure[
                0]  # Return the root of the tree as ADT object.
        except:
            self.ADT.usedIDs = pastusedIDs # If import process fails, set usedIDs back to usedIDs from previous ADT.
            return False # If file not found, then return False.


def createADTInstance(ID: int, type: int, refinement: int, label: str, parent: Optional['ADT.ADT'], root: List[Tuple[bool, Optional['ADT.ADT']]]) -> 'ADT.ADT':
    """
    Creates an object of class `ADT`, with information based on XML import. Needed in XML importation process.

    :param ID: ID number of new `ADT` object.
    :param type: Type of new `ADT` object.
    :param refinement: Refinement of new `ADT` object.
    :param label: Label of new `ADT` object.
    :param parent: Parent node of new `ADT` object.
    :param root: Root node of ADT to which new `ADT` object is added.
    :return: `ADT` object: returned after constructor is called to create a new node.

    Note: *Helper function, users should not call this function themselves. Used in member function: `importXML(...)`*
    """
    return ADT.ADT(ID, type, refinement, label, parent, root)


# Developed by Jayme Hebinck, for bachelor thesis Leiden University.
# This file is part of the ADT Create project, found on: https://github.com/MainJay/ADTPythonLibrary/.

from ADTCreate.MultisetList import MSList
import math
import Levenshtein
import numpy as np

from . import ADT
from . import MultisetList
from typing import List, Tuple, Dict, Union, Optional

class CompADTerms:
    """
    Contains functions regarding ADTerms Comparison operations on `ADT` objects.
    """
    def __init__(self, ADT: 'ADT.ADT') -> None:
        """
        Links `CompADTerms` object to `ADT` object. `CompADTerms` object is initialized when `ADT` object is initialized.
        :param ADT: Links `ADT` object to this `CompADTerms` object, to make sure these two objects are linked correctly.
        """
        self.ADT = ADT

    def generateItemsToCompare(self, ADTermsMS: str) -> 'MultisetList.MSList':
        """
        Generates items to compare based on ADTerms string and put the items in a multiset.

        :param ADTermsMS: ADTerm that should have its items converted into a multiset.
        :return: `MSList` object: multiset list object with all items of ADTerm string.

        Note: *Helper function, users should not call this function themselves. Used in member function: `multisetEquivalenceCheck(...)`.*
        """
        depthLevel = 0  # Depth level is kept to only generate items that are on the surface level.
        inString = False  # Keep track of whether we are working on a string right now.
        itemsToCompare = MSList()  # Keep track of all separate items.
        entry = ""  # Initialize empty entry in itemsToCompare MSList.
        for charIndex in range(len(ADTermsMS)):
            if ADTermsMS[charIndex] == "[" and not inString:
                # If this character is a [ and we are not working on a string, go a level deeper.
                depthLevel += 1
            elif ADTermsMS[charIndex] == "]" and not inString:
                # If this character is a ] and we are not working on a string, go a level undeeper.
                depthLevel -= 1
            elif ADTermsMS[charIndex] == "\"" and depthLevel == 0:
                # If this character is a " and we are on the surface level.
                if not inString:
                    # If a string has not been opened yet, open a string and add this character to the entry.
                    inString = True
                    entry += ADTermsMS[charIndex]
                elif (inString) and (charIndex + 1 < len(ADTermsMS) - 1) and not (
                        ADTermsMS[charIndex + 1] == "," or ADTermsMS[charIndex + 1] == "]"):
                    # If a string has been opened, and this is a " with no , or ] after it, then this " is part of the string.
                    entry += ADTermsMS[charIndex]  # Add this character to the entry.
                elif inString:
                    # If this is a string and the next character is a , or ], this is the end of the string.
                    inString = False
                    entry += ADTermsMS[charIndex]  # Add this character to the entry.
                    itemsToCompare.append(entry)  # Add entry to the itemsToCompare MSList.
                    entry = ""  # Initialize new empty entry.
            elif inString and depthLevel == 0:
                # If we are working on a string and we are on the surface level, add this character to the entry.
                entry += ADTermsMS[charIndex]
            elif (ADTermsMS[charIndex] == "∧" or ADTermsMS[charIndex] == "∨" or ADTermsMS[
                charIndex] == "c") and depthLevel == 0:
                # If the current character is a symbol and we are on the surface level, add it to the entry.
                if ADTermsMS[charIndex] == "c" and ADTermsMS[charIndex + 2] == "[" and (
                        ADTermsMS[charIndex + 3] == "∧" or ADTermsMS[charIndex + 3] == "∨"):
                    # If this symbol is a counter symbol, check if a ∧ or a ∨ symbol follows directly. If so, add these
                    # symbols to the entry.
                    entry += ADTermsMS[charIndex]
                    entry += ADTermsMS[charIndex + 1]
                    entry += ADTermsMS[charIndex + 2]
                    entry += ADTermsMS[charIndex + 3]
                    entry += ADTermsMS[charIndex + 4]
                else:
                    # If this symbol is not a counter symbol, or this counter symbol does not have any direct other
                    # symbols, add only the counter symbol to the entry.
                    entry += ADTermsMS[charIndex]
                    entry += ADTermsMS[charIndex + 1]
                itemsToCompare.append(entry)  # Add entry to the itemsToCompare MSList.
                entry = ""  # Initialize new empty entry.
        return itemsToCompare


    def determineLevenshteinDistance(self, labelMS1: str, MS2: MultisetList.MSList) -> Tuple[bool, Optional[str], Optional[int]]:
        """
        Determines the Levenshtein Distance between a label from the label we want to compare to all elements in the other multiset. Determines if label is within equivalence threshold of any other label from the other multiset, and picks the label that has the shortest Levenshtein Distance and is within the equivalence threshold.

        :param labelMS1: String with label that needs to be compared.
        :param MS2: Multiset with all elements from other ADTerm.
        :return: `bool`, `string`/`None`, `int`/`None`: `True`, element with smallest Levenshtein Distance, Levenshtein Distance if label has been found within equivalence threshold, `False`, `None`, `None` otherwise.

        Note: *Helper function, users should not call this function themselves. Used in member functions: `multisetEquivalenceCheck(...)`, `parentChildCheck(...)`.*
        """
        possibleEntry = []
        for element in MS2:
            # For every element in MS2
            if element[0] == "\"" and element[-1] == "\"":
                longestLabel = max(len(labelMS1) - 2,
                                   len(element) - 2)  # Takes length of biggest label (both -2 to remove the " before and after the label).
                errorAcceptThreshold = math.ceil(longestLabel * self.ADT.LSEquivalenceThreshold)
                # If this element starts and ends with ", we know it is a label.
                if Levenshtein.distance(labelMS1, element) <= errorAcceptThreshold:
                    entry = [element, Levenshtein.distance(labelMS1, element)]
                    possibleEntry.append(entry)
                    # Perform Levenshtein Distance, if within error threshold, return true, the element of MS2 and Levenshtein Distance.
        if not possibleEntry:
            return False, None, None
        smallest = np.Inf
        smallestEntry = None
        for entry in possibleEntry:
            if entry[1] < smallest:
                smallestEntry = entry
                smallest = entry[1]
        return True, smallestEntry[0], smallestEntry[1]
        # If no label found that is within error threshold, return False.



    def multisetEquivalenceCheck(self, ADTermsMS1: 'MultisetList.MSList', ADTermsMS2: 'MultisetList.MSList') -> Tuple[bool, int]:
        """
        Checks whether the two multisets containing elements of ADTerm on a specific level of abstraction, and checks if the these two multisets are equivalent.

        :param ADTermsMS1: ADTerm to compare with.
        :param ADTermsMS2: ADTerm to compare with.
        :return: `bool`, `int`/`None`: `True` and total Levenshtein Distance (`Int`) if equivalent, `False` and `None` otherwise.

        Note: *Helper function, users should not call this function themselves. Used in member functions: `recursiveComparison(...)`, `compareADTerms(...)`.*
        """
        MS1 = self.generateItemsToCompare(ADTermsMS1)  # Generate MS of items to compare of ADTerms1
        MS2 = self.generateItemsToCompare(ADTermsMS2)  # Generate MS of items to compare of ADTerms2
        totalLevenShteinDistance = 0
        for element in MS1:
            MS2Copy = MS2
            if element[0] == "\"" and element[-1] == "\"":
                # If this element in the Multiset is a label.
                resultLS = self.determineLevenshteinDistance(element, MS2Copy)
                if not resultLS[0]:
                    return False, 0
                totalLevenShteinDistance += resultLS[2]
                MS2[MS2.index(resultLS[1])] = element  # Change this entry temporarily in this MS so that the final check
                # will pass with the Levenshtein distance processed.
        if MS1 == MS2:
            # Compare ADTerms1 vs ADTerms2, if equivalent, return True with total Levenshtein Distance so far.
            return True, totalLevenShteinDistance
        # Else, return False
        return False, 0


    def selectCurrentMultisetStr(self, ADTerms: str) -> str:
        """
        Selects elements of ADTerm on a specific level of abstraction and saves them in string format.

        :param ADTerms: ADTerm to select elements of.
        :return: `str`: string of elements of ADTerm.

        Note: *Helper function, users should not call this function themselves. Used in member function: `recursiveComparison(...)`.*
        """
        depthLevel = 0  # Depth level is kept to only select the Multiset parts that are one level deeper.
        MSParts = MSList()  # Keep track of all separate parts of deeper Multisets
        inBetween = False  # Keep track of whether we are in between two separate deeper level Multisets.
        for charIndex in range(len(ADTerms)):
            if ADTerms[charIndex] == "[" and depthLevel == 0:
                # If this character is a [ and we are on the surface level, we determine that this is the start of a new
                # deeper level Multiset. Set inBetween on False, openPos to the next character and depthlevel a level
                # deeper.
                inBetween = False
                openPos = charIndex + 1
                depthLevel += 1
            elif ADTerms[charIndex] == "[" and depthLevel > 0:
                # If this character is a [ and we are already in a deeper level, we are not interested in this
                # Multiset yet. Set depthLevel a level deeper.
                depthLevel += 1
            elif ADTerms[charIndex] == "]" and depthLevel == 1:
                # If this character is a ] and the depthLevel is 1, this is the end of the deeper level Multiset.
                # Depthlevel will now be set back to surface level (0), closePos is set to this character,
                # MSParts saves this deeper Multiset, and we are now potentially inBetween two Multisets, so set
                # inBetween to True.
                depthLevel -= 1
                closePos = charIndex
                MSParts.append(ADTerms[openPos:closePos])
                inBetween = True
            elif ADTerms[charIndex] == "]" and depthLevel > 1:
                # If this character is a ] and depthLevel is higher than 1, we are not interested in this Multiset yet.
                # Set depthLevel a level undeeper.
                depthLevel -= 1
            elif (ADTerms[charIndex] == "," or ADTerms[charIndex] == " ") and depthLevel == 0 and inBetween:
                # If this character is a , or a ' ', and we are on the surface level and in between two Multisets,
                # add this , or ' ' to the MSParts.
                MSParts += ADTerms[charIndex]
            elif (ADTerms[charIndex] != "," or ADTerms[charIndex] != " ") and inBetween:
                # If this character is not a , or a ' ', and we are inBetween two Multisets, we can determine that there
                # will not be a next Multiset. Set inBetween to False.
                inBetween = False
        totalString = "".join(MSParts)  # After all MSParts are saved, join them together in a string.
        return totalString  # Return this deeper most outer Multiset as string.


    def splitMultiset(self, ADTerms: str) -> 'MultisetList.MSList':
        """
        Split given Multiset into different entries (entries are current level (intermediary/leaf) nodes).

        :param ADTerms: ADTerm to select elements of.
        :return: `MSList`: Multiset of elements of ADTerm.

        Note: *Helper function, users should not call this function themselves. Used in member function: `parentChildCheck(...)`.*
        """
        # Split given Multiset into different entries (entries are current level (intermediary/leaf) nodes)
        # Important to note: since spaces between entries are no longer of use, we strip the individual entries before adding
        # them to the entries MSList
        depthLevel = 0
        startPos = 0
        entries = MSList()
        for charIndex in range(len(ADTerms)):
            if ADTerms[charIndex] == "[":
                # When character [ is seen, we are a level deeper.
                depthLevel += 1
            if ADTerms[charIndex] == "]":
                # When character ] is seen, we are a level undeeper.
                depthLevel -= 1
            if ADTerms[charIndex] == "," and depthLevel == 0:
                # When a comma is seen, and we are on the surface level, we determine that this is a separation of an entry.
                endPos = charIndex
                entry = ADTerms[startPos:endPos].strip()
                entries.append(entry)
                startPos = charIndex + 1
        # Last part of the string is also an entry.
        entry = ADTerms[startPos:].strip()
        entries.append(entry)
        return entries


    def getSymbol(self, entryMS: str) -> str:
        """
        Separates symbol from multiset.

        :param entryMS: Multiset which contains multiset.
        :return: `str`: Symbol if there, otherwise `""`.

        Note: *Helper function, users should not call this function themselves. Used in member function: `parentChildCheck(...)`.*
        """
        # Get symbol of current level ADTerms entries. If no symbol (just a label name), return empty string instead.
        for charIndex in range(len(entryMS)):
            if entryMS[charIndex] == "[":
                symbol = entryMS[:charIndex]
                return symbol
        else:
            return ""


    def generateParentChildPairs(self, elementsADTerms: 'MultisetList.MSList', entriesMS: 'MultisetList.MSList') -> MultisetList.MSList:
        """
        Generate pairs of deeper level ADTerms entries and current level ADTerms symbols.

        :param elementsADTerms: Contains symbols of ADTerm.
        :param entriesMS: Contains labels of multiset.
        :return: `MSList`: multiset with pairs between labels and symbols.

        Note: *Helper function, users should not call this function themselves. Used in member function: `parentChildCheck(...)`.*
        """
        # Generate pairs of deeper level ADTerms entries and current level ADTerms symbols.
        pairs = MSList()
        for element in elementsADTerms:
            for entry in range(len(entriesMS)):
                if element in entriesMS[entry]:
                    # If deeper level part of ADTerms is found in current level ADTerms, save the symbol it must be
                    # connected to from current level ADTerms, and make a pair containing a list of all labels from the
                    # deeper level ADTerms and the current level ADTerms symbol that they have to be present under.
                    symbol = self.getSymbol(entriesMS[entry])
                    pair = (self.generateNodeLabels(element), symbol)
                    pairs.append(pair)  # Append this pair to MSList of pairs.
        # All pairs are added in the pairs MSList. Return this MSList.
        return pairs


    def generateNodeLabels(self, ADTerms: str) -> 'MultisetList.MSList':
        """
        Generate a multiset with the labels in this ADTerms.

        :param ADTerms: ADTerm string which contains the labels.
        :return: `MSList`: multiset with labels of the node.

        Note: *Helper function, users should not call this function themselves. Used in member function: `generateParentChildPairs(...)`.*
        """
        # Generate a MSList with the labels in this ADTerms.
        labels = MSList()
        label = ""
        inString = False
        for charIndex in range(len(ADTerms)):
            # A label is present between "".
            if ADTerms[charIndex] == "\"" and not inString:
                label += ADTerms[charIndex]
                inString = True
            elif ADTerms[charIndex] == "\"" and inString:
                label += ADTerms[charIndex]
                labels.append(label)
                label = ""
                inString = False
            elif inString:
                label += ADTerms[charIndex]
        # Return the MSList with all labels.
        return labels


    def parentChildCheck(self, ADTerms1: str, ADTerms2: str, oldADTerms1: str, oldADTerms2: str) -> bool:
        """
        Generate entries for Multisets of deeper level ADTerms and current level ADTerms. Generate pairs of deeper level ADTerms and current level ADTerms symbols. Parent-Child relationship is established. Compare pairs of multisets. If one pair is not found in the other pair, parent-child relationship is not maintained.

        :param ADTerms1: ADTerm of current level.
        :param ADTerms2: ADTerm of current level.
        :param oldADTerms1: ADTerm of deeper level.
        :param oldADTerms2: ADTerm of deeper level.
        :return: `bool`: `True` if parent-child relationships are equivalent, `False` otherwise.

        Note: *Helper function, users should not call this function themselves. Used in member functions: `recursiveComparison(...)`, `compareADTerms(...)`.*
        """
        # Generate entries for Multisets of deeper level ADTerms and current level ADTerms
        entriesMS1 = self.splitMultiset(ADTerms1)
        entriesMS2 = self.splitMultiset(ADTerms2)
        splitElementsADTerms1 = self.splitMultiset(oldADTerms1)
        splitElementsADTerms2 = self.splitMultiset(oldADTerms2)

        # Generate pairs of deeper level ADTerms and current level ADTerms symbols. Parent-Child relationship is established.
        pairsMS1 = self.generateParentChildPairs(splitElementsADTerms1, entriesMS1)
        pairsMS2 = self.generateParentChildPairs(splitElementsADTerms2, entriesMS2)

        # Compare pairs of multisets. If one pair is not found in the other pair, parent-child relationship is not maintained.
        # This means that these two ADTerms are not equivalent.
        # Important note: we are only comparison the labels present in the deeper level ADTerms, not the whole ADTerms
        # structure. Labels is enough considering the ADTerms structure is compared in deeper checks.
        # This check is only for parent-child relationship, where it is important that the labels are part of the
        # deeper level ADTerms that are connected to the right symbol of the current level ADTerms.
        labelsFound = False  # If all labels are found in a given entry of the other Multiset.
        for entryMS1 in pairsMS1:
            # For an entry in the pairs of Multiset 1.
            if entryMS1 not in pairsMS2:
                # If this entry is not exactly in the pairs of Multiset 2.
                for entryMS2 in pairsMS2:
                    # Check for every label in every entry of Multiset 2 if it is within error threshold of first label
                    # in the entry of Multiset 1.
                    resultLS = self.determineLevenshteinDistance(entryMS1[0][0], entryMS2[0])
                    if resultLS[1]:
                        # If the first label of the entry of Multiset 1 finds a label in the entry of Multiset 2 within
                        # error threshold, we test all other labels of Multiset 1 on this entry of Multiset 2.
                        labelsFound = True
                        for label in entryMS1[0]:
                            if label != entryMS1[0][0]:
                                # We skip the first label, since we already determined that one was found.
                                labelsResultLS = self.determineLevenshteinDistance(label, entryMS2[0])
                                if not labelsResultLS[1]:
                                    # If any of the other labels in the entry of Multiset 1 are not within the error
                                    # threshold of the labels in the entry of Multiset 2, these pairs are not equivalent.
                                    labelsFound = False
                                    break
                        if labelsFound:
                            # If all labels in the entry of Multiset 1 are within the error threshold of (or equal to)
                            # all the labels in the entry of Multiset 2, these pairs are equivalent.
                            break
                        # break
                if not labelsFound:
                    # If one or more of the labels in the entry of Multiset 1 are not within the error threshold of
                    # the labels in the entry of Multiset 2, these pairs are not equivalent.
                    return False
                labelsFound = False  # After this check, reset labelsFound boolean back to False for next check.
        # If no proof is found against equivalence, return True.
        return True


    def recursiveComparison(self, ADTerms1Sub: str, ADTerms2Sub: str, continueCheck: bool, equivalence: bool) -> Tuple[bool, str, str, int]:
        """
        Recursively compares two ADTerms by performing the top-down method and bottom-up method.

        :param ADTerms1Sub: String of part of ADTerm that needs to be compared.
        :param ADTerms2Sub: String of part of ADTerm that needs to be compared.
        :param continueCheck: Boolean that determines if we continue the check, or not when equivalence is disproved.
        :param equivalence: Boolean that shows if the ADTerm is equivalent or not (within equivalence threshold).
        :return: `bool`, `str`, `str`, `int`: `True`, string of part of ADTerm, string of part of ADTerm and Levenshtein Distance when equivalent till this point, `False`, string of part of ADTerm, string of part of ADTerm and `0` when no longer equivalent.

        Note: *Helper function, users should not call this function themselves. Used in member functions: `recursiveComparison(...)`, `compareADTerms(...)`.*
        """
        if continueCheck:
            # As long as we have not found a problem, continue checking.
            ADTerms1Selected = self.selectCurrentMultisetStr(ADTerms1Sub)  # Get deeper level ADTerms
            ADTerms2Selected = self.selectCurrentMultisetStr(ADTerms2Sub)  # Get deeper level ADTerms
            ADTerms1Sub, ADTerms2Sub = ADTerms1Selected, ADTerms2Selected  # Save current level ADTerms
            totalLevenshteinDistance = 0
            if ADTerms1Sub == "" and ADTerms2Sub == "":
                # If both ADTerms are empty, check has finished successfully. Return with equivalence == True.
                return equivalence, ADTerms1Sub, ADTerms2Sub, 0
            resultMSCheck = self.multisetEquivalenceCheck(ADTerms1Sub, ADTerms2Sub)
            if not resultMSCheck[0]:
                # Not equivalent, stop the check and set equivalence to False.
                equivalence = False
                return equivalence, ADTerms1Sub, ADTerms2Sub, 0

            # If not finished yet, perform check a level deeper.
            equivalence, deeperADTerms1Sub, deeperADTerms2Sub, totalLevenshteinDistance = self.recursiveComparison(
                ADTerms1Sub, ADTerms2Sub, continueCheck, equivalence)
        # After all levels are checked, go back up out of recursion and check parent-child relationship.
        if equivalence:
            parentChildCheck = self.parentChildCheck(ADTerms1Sub, ADTerms2Sub, deeperADTerms1Sub, deeperADTerms2Sub)
        else:
            # If already found that at this point the two ADTerms are not equivalent, set to False.
            parentChildCheck = False

        # Check parent-child relationship with current ADTerms substrings and the deeper level ADTerms substrings. Match them.
        if not parentChildCheck:
            # If parentChildCheck has failed, set equivalence to False.
            equivalence = False

        # Add new Levenshtein Distance to total.
        totalLevenshteinDistance += resultMSCheck[1]

        # Return with four variables: equivalence, the two needed subADTerms to further compare and the total Levenshtein Distance so far.
        return equivalence, ADTerms1Sub, ADTerms2Sub, totalLevenshteinDistance

    def compareADTerms(self, ADTerms1: str, ADTerms2: str) -> Tuple[bool, int]:
        """
        Determines whether two ADTerms are equivalent to each other by comparing the two. Includes check with equivalence threshold using the Levenshtein Distance.

        :param ADTerms1: ADTerm to compare with.
        :param ADTerms2: ADTerm to compare with.
        :return: `bool`, `int`/`None`: `True` and total Levenshtein Distance (`int`) if equivalent, `False` and `None` otherwise.

        Note: *This function is called from the `ADT` class using `compareADTerms(...)`, and should therefore not be called from the `CompADTerms` class.*
        """
        # Set variables continueCheck and equivalence to True.
        # We determine equivalence by finding proof that these two ADTerms are NOT equivalent, not the other way around.
        continueCheck = True
        equivalence = True
        totalLevenshteinDistance = 0

        # If first character of both ADTerms is a symbol, we firstly determine equivalence of these symbols before we
        # start a recursive comparison of Multisets within the ADTerms.
        if ADTerms1[0] != "[" and ADTerms2[0] != "[":
            resultMSCheck = self.multisetEquivalenceCheck(ADTerms1, ADTerms2)
            if not resultMSCheck[0]:
                # If at any time proof is found that the given ADTerms are not equivalent, disable continueCheck.
                continueCheck = False
                return False, 0
            totalLevenshteinDistance += resultMSCheck[1]
        # If first character is not a symbol, or symbols are equivalent, start recursive comparison.
        result = self.recursiveComparison(ADTerms1, ADTerms2, continueCheck, equivalence)

        # After recursive comparison, check whether the comparison resulted in equivalence or not.
        totalLevenshteinDistance += result[3]
        if not result[0]:
            return False, 0
        else:
           return True, totalLevenshteinDistance

    def changeLevenshteinErrorMargin(self, newValue: float) -> bool:
        """
        Changes Levenshtein Distance equivalence threshold value if new value is between 0.0 and 1.0.

        :param newValue: New value of Levenshtein Distance equivalence threshold.
        :return: `bool`: `True` if value changed successfully, `False` otherwise.

        Note: *This function is called from the `ADT` class using `changeLevenshteinErrorMargin(...)`, and should therefore not be called from the `CompADTerms` class.*
        """
        # Changes Levenshtein Distance equivalence threshold with a new value. This value must be in [0, 1].
        if float(newValue) >= 0.0 or float(newValue) <= 1.0:
            self.ADT.LSEquivalenceThreshold = float(newValue)
            return True
        return False


# Developed by Jayme Hebinck, for bachelor thesis Leiden University.
# This file is part of the ADT Create project, found on: https://github.com/MainJay/ADTPythonLibrary/.

class MSList(list):
    """
    Implementation of multisets, based on `list`. `list` is edited so that the order of elements do not matter.
    """
    # This class inherits the built-in list class of Python. It then changes the __eq__ function, so that comparison is
    # not based on the order of elements in the list, but rather on if this item even is in this MSList.
    # A list like this, where order does not matter, is called a Multi-Set. Therefore, this class is called the
    # MSList.
    def __eq__(self, other: 'MSList') -> bool:
        """
        This equal function has been modified slightly, so that order of elements does not matter when two objects of `MSList` are compared.

        :param other: Other multiset that this multiset is compared to.
        :return: `bool`: `True` when multisets are equal, `False` otherwise.
        """
        for item in self:
            if item not in other:
                return False
        return True






from setuptools import setup, find_packages

setup(
    name="ADTCreate",
    version="1.0.0rc2",
    package_dir={'': 'src'},
    packages=find_packages(where='src'), 
    install_requires=[
        "Levenshtein>=0.23.0",
        "numpy>=1.21.6",
    ],
    python_requires=">=3.7",
    author="Jayme Hebinck",
    author_email="jayme.hebinck@live.nl",
    description="ADT Create provides a general method to create and customize Attack-Defense Trees (ADTs), allowing users to visualize and analyse security scenarios. A user is able to create and customize ADTs from scratch, upload/download ADTs in XML format, generate ADTerms based on ADTs, compare ADTerms to determine equivalence using an equivalence threshold using the Levenshtein Distance, determine satisfiability of an ADT and generate statistics of an ADT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/MainJay/ADTPythonLibrary/",
    project_urls={
        'Tracker': "https://github.com/MainJay/ADTPythonLibrary/issues",
    },
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)